#include "SSP.h"

StaticBatchSSP::StaticBatchSSP(FABMAS_SB_ToolGroup* tg, string name) : FABMAS_SB_SSP(tg, name) 
{
}

void StaticBatchSSP::initialize_problem() {
  FABMAS_SB_SSP::initialize_problem();

  // calculate batchable tasks
  list<FABMAS_SB_Task*>::const_iterator i = schedulable_tasks.begin();
  while (i != schedulable_tasks.end()) {
    FABMAS_SB_Task* task = *i;
    tasks_grouped_by_family[task->get_setup()].push_back(task);
    i++;
  }
}

void StaticBatchSSP::solve_problem() {
  while (!schedulable_tasks.empty()) {
    FABMAS_SB_Tool* tool = toolgroup->calculate_next_avail_tool();

    // choose a batch
    FABMAS_SB_Batch choosed_batch = choose_batch(tool);
    FABMAS_SB_Date batch_avail_time = choosed_batch.calculate_avail_time();

    // consider possible switching of machines
    if (toolgroup->get_scheduler()->config->get_bool("try_to_switch_machine")) {
      tool = toolgroup->calculate_latest_avail_tool(batch_avail_time, choosed_batch.get_setup());
    }

    if (!choosed_batch.tasks.empty()) {
      // finally add the task to the schedule of tool tool.
      tool->add_to_schedule(choosed_batch);
		
      filter_tasks(tasks_grouped_by_family[choosed_batch.get_setup()], choosed_batch.tasks);
      if (tasks_grouped_by_family[choosed_batch.get_setup()].empty()) 
	tasks_grouped_by_family.erase(choosed_batch.get_setup());

      filter_tasks(schedulable_tasks, choosed_batch.tasks);
    } else {
      list<FABMAS_SB_Task*>::iterator i = schedulable_tasks.begin();
      while (i != schedulable_tasks.end()) {
	FABMAS_SB_Task* task = *i;
	task->mark_as_unschedulable();
	i++;
      }

      tasks_grouped_by_family.clear();
      schedulable_tasks.clear();
    }
  }
}

FABMAS_SB_Batch StaticBatchSSP::choose_batch(FABMAS_SB_Tool* tool) {	
  // choose the best batch from each family
  FABMAS_SB_Batch choosed_batch;
  map<FABMAS_SB_Setup*, list<FABMAS_SB_Task*> >::iterator k = tasks_grouped_by_family.begin();
  while (k != tasks_grouped_by_family.end()) {
    list<FABMAS_SB_Task*>& tasks = (*k).second;

    list<FABMAS_SB_Task*> avail_tasks;
    list<FABMAS_SB_Task*>::iterator i = tasks.begin();
    while (i != tasks.end()) {
      FABMAS_SB_Task* task = *i;
      if (task->is_available()) avail_tasks.push_back(task);

      i++;
    }

    FABMAS_SB_Batch batch = build_batch(avail_tasks, tool);
    if (better_batch(batch, choosed_batch)) choosed_batch = batch;

    k++;
  }

  return choosed_batch;
}

FABMAS_SB_Batch StaticBatchSSP::build_batch(const list<FABMAS_SB_Task*>& tasks, FABMAS_SB_Tool* tool) {
  FABMAS_SB_Batch choosed_batch; FABMAS_SB_Batch batch;

  // get all tasks, which are already available
  FABMAS_SB_Task* task = get_best_task_available_at_current_time(tool, tool->get_avail_time(), tasks);
  while ((batch.tasks.size() < tool->get_batch_max()) && (task != 0)) {
    task->set_scheduled_tool(tool); batch.tasks.insert(task);
    task = get_best_task_available_at_current_time(tool, tool->get_avail_time(), tasks);
  }

  // let see if the batch can be filled in another way
  if (batch.tasks.size() < tool->get_batch_max()) {
    //fill the batch with the next ready tasks until it reaches the minimum batch size
    bool no_task_found = false;
    while ((batch.tasks.size() < tool->get_batch_min()) && (!no_task_found)) {
      task = get_earliest_available_task_at_current_time(tool, tool->get_avail_time(), tasks);

      if (task == 0) {
	no_task_found = true;
      } else {
	task->set_scheduled_tool(tool);
	batch.tasks.insert(task);
      }
    }

    if (!no_task_found) {
      FABMAS_SB_Date current_time = max(tool->get_avail_time(), batch.calculate_avail_time());
      batch.set_dynamic_priority(calculate_objective(tool, current_time, batch));	
      choosed_batch = batch;
      toolgroup->get_graph()->log_trace("evaluating batch " + choosed_batch.to_string());

      // now see if we can get better objectives
      while ((batch.tasks.size() < tool->get_batch_max()) && (!no_task_found)) {
	task = get_earliest_available_task_at_current_time(tool, batch.calculate_avail_time(), tasks);

	if (task == 0) {
	  no_task_found = true;
	} else {
	  task->set_scheduled_tool(tool); batch.tasks.insert(task);
	  batch.set_dynamic_priority(calculate_objective(tool, current_time, batch));

	  toolgroup->get_graph()->log_trace("evaluating batch " + batch.to_string());

	  if (batch.calculate_avail_time() <= choosed_batch.calculate_avail_time()) {
	    choosed_batch = batch;
	  } else {
	    if (better_batch(batch, choosed_batch)) choosed_batch = batch;
	  }
	}
      }
    }
  } else {
    toolgroup->get_graph()->log_trace("evaluating batch " + batch.to_string());
    choosed_batch = batch;
  }

  list<FABMAS_SB_Task*>::const_iterator i = tasks.begin();
  while (i != tasks.end()) {
    FABMAS_SB_Task* task = *i;
    task->set_scheduled_tool(0);
    i++;
  }

  batch.set_scheduled_tool(0); choosed_batch.set_scheduled_tool(0);
  return choosed_batch;
}
